#ifndef GAIM_PLUGINS
#	define GAIM_PLUGINS
#endif
#include "internal.h"

#include "signals.h"
#include "util.h"
#include "sound.h"
#include "conversation.h"
#include "privacy.h"

#include "gtkplugin.h"
#include "gtkutils.h"
#include "version.h"
#include "blist.h"
#include "cmds.h"

#include "ignorance.h"
#include "ignorance_violation.h"
#include "interface.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <signal.h>
#include <errno.h>
#include <unistd.h>

#ifndef _WIN32
#	include <sys/types.h>
#	include <sys/select.h>
#	include <fcntl.h>
#	include <sys/stat.h>
#endif

#define IGNORANCE_PLUGIN_ID "Ignorance"
#define SHORTDESC "Ignorance filter"

#ifndef IGNORANCE_VERSION
#	define IGNORANCE_VERSION PACKAGE_VERSION
#endif
#ifndef IGNORANCE_CONFDIR
#	define IGNORANCE_CONFDIR gaim_user_dir()
#endif

#define EXEC_TIMEOUT 10

#define GREATER(x,y) ((x)?((x)>(y)):(y))

#if ((GAIM_MAJOR_VERSION) < 2)
#define GAIM_CONV_TYPE_CHAT GAIM_CONV_CHAT
#define GAIM_CONV_TYPE_IM GAIM_CONV_IM
#endif

//list of tolerance levels
static GPtrArray *levels;

//list of commands for unregistering at unload time
static GList	*ignorance_commands;

//
// EXPORTED FUNCTIONS
//
static gboolean plugin_load (GaimPlugin * plugin);
static gboolean plugin_unload(GaimPlugin *plugin);
static void init_plugin (GaimPlugin * plugin);
static GtkWidget* get_config_frame(GaimPlugin *plugin);


//
// PRIVATE FUNCTIONS
//
static gboolean	apply_rule(GaimConversation *conv, GaimAccount *account,const GString *username, const GString *text, gint flags);
static gboolean	ignorance_user_match(ignorance_level *il, const GString *username);
static ignorance_level*		ignorance_get_user_level(const GString *username);
static ignorance_level* ignorance_get_default_level();
static gboolean	ignorance_place_user_name(const GString *level_name,const GString *username);
static void ignorance_print_user_found(GaimConversation *conv, const gchar *username, const gchar *levelname);
static void ignorance_print_user_count(GaimConversation *conv, const gchar *regex, const gchar *levelname, gint count);
static gboolean substitute (GaimConversation *conv,GaimAccount * account, const gchar *sender, gchar **message, gint msgflags);			    
static gboolean load_conf();
static gint buf_get_line(gchar *ibuf, gchar **buf, gint *position, gint len);
static gboolean chat_cb(GaimAccount *account,gchar **sender, gchar **buffer,GaimConversation *chat, void *data);
static gboolean im_cb(GaimAccount *account,gchar **sender, gchar **buffer,gint *flags, void *data);
static gboolean chat_joinleave_cb(GaimConversation *conv, const gchar *name, void *data);
static gint chat_invited_cb(GaimAccount *account,const gchar *inviter, const gchar *chat, const gchar *invite_message, const GHashTable *components, void *data);
static void buddy_added_cb(GaimBuddy *buddy, gpointer data);
static void buddy_removed_cb(GaimBuddy *buddy, gpointer data);

static gboolean import_zinc_list();
static gboolean import_gaim_list();

static void gaim_buddy_add(gpointer key, gpointer value, gpointer user_data);
static GaimCmdRet	command_callback(GaimConversation *gc,const gchar *cmd,gchar **args,gchar **error,void *data);
static gboolean generate_default_levels();
static gchar* yahoo_strip_tattoo(gchar *origmessage);
static GaimCmdRet command_mv(GaimConversation *gc, const gchar **args);
static GaimCmdRet command_ls(GaimConversation *gc, const gchar **args);
static GaimCmdRet command_wl(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_bl(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_rm(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_igwho(GaimConversation *gc, const gchar **args);
static GaimCmdRet command_ig(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_msg(GaimConversation *c, const gchar *cmd, gchar **args, gchar **error, void *data);

#ifdef HAVE_REGEX_H
static GaimCmdRet command_ri(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_rf(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_rin(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_rfn(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_grep(GaimConversation *gc, const gchar **args);
#endif
#ifndef _WIN32
static int read_nonblock(int fd,unsigned long len,unsigned long timeout_usec,GString *inp);
static int handle_exec_command (const gchar *command, GString *result, unsigned long maxlen);
#endif

static gboolean ignorance_wl_user(GaimConversation *gc, const gchar *username, const gchar *actual_levelname);
static gboolean ignorance_bl_user(GaimConversation *gc, const gchar *username, const gchar *actual_levelname);
static gboolean ignorance_rm_user(GaimConversation *gc, const gchar *username);



static gboolean
plugin_load (GaimPlugin * plugin)
{
	void *conv_handle = gaim_conversations_get_handle ();
	void *blist_handle = gaim_blist_get_handle();
	GaimCmdId	gcid;
  
	gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Loading plugin %s\n", IGNORANCE_PLUGIN_ID);
	
	load_conf();

	//connect signals
	gaim_signal_connect (conv_handle, "receiving-im-msg",
				plugin, GAIM_CALLBACK (im_cb), NULL);
	gaim_signal_connect (conv_handle, "receiving-chat-msg",
				plugin, GAIM_CALLBACK (chat_cb), NULL);
	gaim_signal_connect(conv_handle,"chat-buddy-joining",
				plugin,GAIM_CALLBACK(chat_joinleave_cb),NULL);
	gaim_signal_connect(conv_handle,"chat-buddy-leaving",
				plugin,GAIM_CALLBACK(chat_joinleave_cb),NULL);
	gaim_signal_connect(conv_handle,"chat-invited",
				plugin,GAIM_CALLBACK(chat_invited_cb),NULL);
	gaim_signal_connect(blist_handle,"buddy-added",
				plugin,GAIM_CALLBACK(buddy_added_cb),NULL);
	gaim_signal_connect(blist_handle,"buddy-removed",
				plugin,GAIM_CALLBACK(buddy_removed_cb),NULL);
	
	//register commands with gaim core
	gcid=gaim_cmd_register("mv","ww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT,NULL,command_callback,"mv user group:  Move a user to a group",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("ls","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT | GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_callback,"ls group1 group2 ...:  List all users in each given group",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("wl","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT | GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_callback,"wl user1 user2 ...:  Whitelist each given user",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("bl","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT | GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_callback,"bl user1 user2 ...:  Blacklist each given user",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("ig","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT | GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_callback,"ig user1 user2 ...:  Blacklist each given user",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("rm","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT | GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_callback,"rm user1 user2 ...:  Remove each given user from his current group",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("igwho","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM|GAIM_CMD_FLAG_CHAT|GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_callback,"igwho group1 group2 ...: Display users in the current conversation that belong to the given groups",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("msg", "ws", GAIM_CMD_P_PRPL,
	                  GAIM_CMD_FLAG_CHAT | GAIM_CMD_FLAG_PRPL_ONLY,
	                  "prpl-yahoo", command_msg,
	                  _("msg &lt;contact&gt; &lt;message&gt;: Send a message to a contact"), NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);

#ifdef HAVE_REGEX_H
	gcid=gaim_cmd_register("ri","w",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT ,NULL,command_callback,"ri expression:  Add a regex autoignore rule with the given expression to the Default group",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("rf","w",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT ,NULL,command_callback,"rf expression:  Add a regex filter rule with the given expression to the Default group",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("rin","w",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT ,NULL,command_callback,"rin expression:  Add a regex username autoignore rule with the given expression to the Default group",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("rfn","w",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT ,NULL,command_callback,"rfn expression:  Add a regex username filter rule with the given expression to the Default group",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("grep","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM|GAIM_CMD_FLAG_CHAT|GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_callback,"grep [-c] userexp group1 group2 ...:  Search each given group's usernames for matches to the regex userexp",NULL);
	ignorance_commands=g_list_prepend(ignorance_commands,&gcid);
#endif
	
  return TRUE;
}


static gboolean
plugin_unload(GaimPlugin *plugin){
	GList	*cursor;
	GaimCmdId	gcid;
	gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Unloading plugin %s\n", IGNORANCE_PLUGIN_ID);
	save_conf();
	
	for(cursor=ignorance_commands;cursor;cursor=cursor->next){
		gcid=*((GaimCmdId*)(cursor->data));
		gaim_cmd_unregister(gcid);
	}//for
	
	g_list_free(ignorance_commands);
	
	return TRUE;
}//plugin_unload


//Passes back plugin's configuration panel
static GtkWidget *
get_config_frame(GaimPlugin *plugin)
{
	return create_uiinfo(levels);
}//get_config_frame


static GaimGtkPluginUiInfo ui_info =
{
	get_config_frame
};


static GaimPluginInfo info =
{
	GAIM_PLUGIN_MAGIC,
	GAIM_MAJOR_VERSION,
	GAIM_MINOR_VERSION,
	GAIM_PLUGIN_STANDARD,
	GAIM_GTK_PLUGIN_TYPE,
	0,
	NULL,
	GAIM_PRIORITY_DEFAULT,
	IGNORANCE_PLUGIN_ID,
	N_(SHORTDESC),
	IGNORANCE_VERSION,
	N_("Allows you to manage lists of users with various levels of allowable activity."),
	N_("Allows you to manage lists of users with various levels of allowable activity."),
	"Levi Bard <taktaktaktaktaktaktaktaktaktak@gmail.com>",
	"http://ignorance.sourceforge.net",
	plugin_load,
	plugin_unload,
	NULL,
	&ui_info,
	NULL,
	NULL,
	NULL
};


//what is the difference between this and plugin_load?!
static void
init_plugin (GaimPlugin * plugin)
{
}//init_plugin

GAIM_INIT_PLUGIN (command, init_plugin, info)
 

//
// Private functions
//

//callback for received chat messages
static gboolean chat_cb(GaimAccount *account,gchar **sender, gchar **buffer,
					 	GaimConversation *chat, void *data){
	return substitute(chat,account,*sender,buffer, IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_USER);
}//chat_cb


//callback for received IMs
static gboolean im_cb(GaimAccount *account,gchar **sender, gchar **buffer,
					 	gint *flags, void *data){
	GaimConversation *gc;
	

//Damn you, API changes!
#if ((GAIM_MAJOR_VERSION) == 2)
	gc=gaim_find_conversation_with_account(GAIM_CONV_TYPE_IM,*sender,account);
#else
	gc=gaim_find_conversation(*sender);
#endif

	return substitute(gc,account,*sender,buffer, IGNORANCE_APPLY_IM | IGNORANCE_APPLY_USER);
}//im_cb


//callback for enter/leave messages
static gboolean chat_joinleave_cb(GaimConversation *conv, const gchar *name, void *data){
	gchar *message=g_strdup(name);
	gboolean rv=substitute(conv,conv->account,name,&message,IGNORANCE_APPLY_ENTERLEAVE | IGNORANCE_APPLY_USER);
	g_free(message);
	
	return rv;
}//chat_joinleave_cb


//callback for chat invites
static gint chat_invited_cb(GaimAccount *account,const gchar *inviter, const gchar *chat, const gchar *invite_message, const GHashTable *components, void *data){
	gchar *message=g_strdup(invite_message);
	gboolean rv;
	gint invite_ask=0;
	GaimConversation *gc;
	g_free(message);

//Damn you, API changes!
#if ((GAIM_MAJOR_VERSION) == 2)
	gc=gaim_find_conversation_with_account(GAIM_CONV_TYPE_IM,inviter,account);
#else
	gc=gaim_find_conversation(inviter);
#endif

	rv=substitute(gc,account,inviter,&message,IGNORANCE_APPLY_INVITE | IGNORANCE_APPLY_USER | IGNORANCE_APPLY_CHAT);

	//Auto-deny invitation 
	if(rv){
		invite_ask=-1;
	}//if

	return invite_ask;
}//chat_invited_cb


static void buddy_added_cb(GaimBuddy *buddy, gpointer data){
	GString	*wlname=g_string_new("WL");
	ignorance_level	*wl=ignorance_get_level_name(wlname);

	gaim_debug(GAIM_DEBUG_INFO,"ignorance","Caught buddy-added for %s%s\n",
								buddy->account->protocol_id,
								buddy->name);

	gaim_buddy_add(NULL,buddy,wl);

	ignorance_level_free(wl);
	g_string_free(wlname,TRUE);
}//buddy_added_cb


static void buddy_removed_cb(GaimBuddy *buddy, gpointer data){
	GString *tmp=g_string_new(buddy->account->protocol_id);
	GaimConversation *conv=NULL;

//Damn you, API changes!
#if ((GAIM_MAJOR_VERSION) == 2)
	conv=gaim_find_conversation_with_account(GAIM_CONV_TYPE_IM,tmp->str,buddy->account);
#else
	conv=gaim_find_conversation(tmp->str);
#endif

	gaim_debug(GAIM_DEBUG_INFO,"ignorance","Caught buddy-removed for %s%s\n",
								buddy->account->protocol_id,
								buddy->name);

	g_string_append(tmp,gaim_normalize(buddy->account,buddy->name));

	ignorance_rm_user(conv,tmp->str);

	g_string_free(tmp,TRUE);
}//buddy_removed_cb


//switching callback for all Ignorance commands
static GaimCmdRet	command_callback(GaimConversation *gc,const gchar *cmd,gchar **args,gchar **error,void *data){
	GaimCmdRet	rv=GAIM_CMD_RET_CONTINUE;
	
	switch(*cmd){
		case 'm':
			if('v'==cmd[1]){
				rv=command_mv(gc,(const gchar**)args);
			}// "/mv"
			break;
		case 'l':
			if('s'==cmd[1]){
				rv=command_ls(gc,(const gchar**)args);
			}// "/ls"
			break;
		case 'w':
			if('l'==cmd[1]){
				rv=command_wl(gc,(const gchar**)args);
			}// "/wl"
			break;
		case 'b':
			if('l'==cmd[1]){
				rv=command_bl(gc,(const gchar**)args);
			}// "/bl"
			break;
		case 'r':
			if('m'==cmd[1]){
				rv=command_rm(gc,(const gchar**)args);
			}// "/rm"
#ifdef HAVE_REGEX_H
			else if('f'==cmd[1]){
				if('n'==cmd[2]){
					rv=command_rfn(gc,(const gchar**)args);
				}else{
					rv=command_rf(gc,(const gchar**)args);
				}// "/rfn"
			}// "/rf"
			else if('i'==cmd[1]){
				if('n'==cmd[2]){
					rv=command_rin(gc,(const gchar**)args);
				}else{
					rv=command_ri(gc,(const gchar**)args);
				}// "/rin"
			}// "/ri"
			break;
		case 'g':
			if(cmd==strstr(cmd,"grep")){
				rv=command_grep(gc,(const gchar**)args);
			}// "/grep"
			break;
#endif
		case 'i':
			if(cmd==strstr(cmd,"igwho")){
				rv=command_igwho(gc,(const gchar**)args);
			}// "/igwho"
			else if(cmd==strstr(cmd,"ig")){
				rv=command_ig(gc,(const gchar**)args);
			}// "/ig"
			break;
	}//switch
	
	return rv;
}//command_callback


//mv username userlevel
static GaimCmdRet command_mv(GaimConversation *gc, const gchar **args){
	GString	*username,
			*levelname;
	gchar	*msgbuf;
	
	//check for proper # args
	if(!args[0]){
		return GAIM_CMD_RET_CONTINUE;
	}else if(!args[1]){
		return GAIM_CMD_RET_CONTINUE;
	}//if
				
	username=g_string_new(gc->account->protocol_id);
	g_string_append(username,args[0]);
	
	levelname=g_string_new(args[1]);
				

	if(strstr(levelname->str,"WL")){
		ignorance_wl_user(gc,args[0],levelname->str);
	}else if(strstr(levelname->str,"BL")){
		ignorance_bl_user(gc,args[0],levelname->str);
	}//if WL or BL
	else{
		msgbuf=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));

		if(ignorance_place_user_name(levelname,username)){
			snprintf(msgbuf,BUFSIZ,"Assigned user %s to %s",username->str,levelname->str);
		}else{
			snprintf(msgbuf,BUFSIZ,"Unable to assign user %s to %s - may already be there",username->str,levelname->str);
		}//if
				
		gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));

		g_free(msgbuf);
	}//regular mv
			
	g_string_free(username,TRUE);
	g_string_free(levelname,TRUE);

	return GAIM_CMD_RET_OK;
}//command_mv


// Lists the users of given groups
// /ls level1, level2, ...
static GaimCmdRet command_ls(GaimConversation *gc, const gchar **args){
	GString	*levelname,
			*username;
	ignorance_level	*level;
	GList	*cursor;
	gchar	*msgbuf;
	gint	i;
	
	if(!args[0]){
		return GAIM_CMD_RET_CONTINUE;
	}//if
				
	levelname=g_string_new("");
				
	for(i=0;args[i];++i){
		g_string_assign(levelname,args[i]);
		level=ignorance_get_level_name(levelname);
		if(level){
			g_string_append(levelname,":");
			gaim_conversation_write(gc,NULL,levelname->str,GAIM_MESSAGE_NO_LOG,time(NULL));
			for(cursor=g_list_first(level->denizens);cursor;cursor=g_list_next(cursor)){
				username=(GString*)(cursor->data);
				gaim_conversation_write(gc,NULL,username->str,GAIM_MESSAGE_NO_LOG,time(NULL));
			}//for
		}else{
			msgbuf=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));
			snprintf(msgbuf,BUFSIZ,"No such group: %s",levelname->str);
			gaim_conversation_write(gc,NULL,msgbuf,GAIM_MESSAGE_NO_LOG,time(NULL));
			g_free(msgbuf);
		}//if
	}//for
				
	g_string_free(levelname,TRUE);
	return GAIM_CMD_RET_OK;
}//command_ls


// Assigns users to the WL level
// /wl user1, user2, ...
static GaimCmdRet command_wl(GaimConversation *gc,const gchar **args){
	gint	i;
				
	for(i=0;args[i];++i){
		ignorance_wl_user(gc,args[i],"WL");
	}//for
				
	return GAIM_CMD_RET_OK;
}//command_wl


// Assigns given users to the BL level
// /bl user1, user2, ...
static GaimCmdRet command_bl(GaimConversation *gc,const gchar **args){
	gint	i;
	
	for(i=0;args[i];++i){
		ignorance_bl_user(gc,args[i],"BL");
	}//for
				
	return GAIM_CMD_RET_OK;
}//command_bl


// Assigns given users to the BL level
// /ig user1, user2, ...
static GaimCmdRet command_ig(GaimConversation *gc,const gchar **args){
	return command_bl(gc,args);
}//command_ig


// Removes given users from their currently assigned level
// (effectively moves them to the default level)
// /rm user1, user2, ...
static GaimCmdRet command_rm(GaimConversation *gc, const gchar **args){
	gint	i;
	GString *tmp=g_string_new("");
				
	for(i=0;args[i];++i){
		g_string_assign(tmp,gc->account->protocol_id);
		g_string_append(tmp,args[i]);
		ignorance_rm_user(gc,tmp->str);
	}//for

	g_string_free(tmp,TRUE);
				
	return GAIM_CMD_RET_OK;
}//command_rm


#ifdef HAVE_REGEX_H
// Regex ignore - adds an ignore regex 
// for IM and chat messages to the Default level
// /ri expression
static GaimCmdRet	command_ri(GaimConversation *gc, const gchar **args){
	gchar	msgbuf[BUFSIZ];
	ignorance_level	*level;
	ignorance_rule	*rule;
	GString	*rulename;
	
	if(!(*args)){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	level=ignorance_get_default_level();
	
	if(!level){
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to get default level\n");
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	rulename=g_string_new(*args);
	
	rule=ignorance_rule_newp(rulename,IGNORANCE_RULE_REGEX,*args,IGNORANCE_FLAG_IGNORE,IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_IM,TRUE,NULL,NULL,NULL);
	if(ignorance_level_add_rule(level,rule)){
		snprintf(msgbuf,BUFSIZ,"Successfully added ignore regex %s",*args);
	}else{
		snprintf(msgbuf,BUFSIZ,"Unable to add ignore regex %s",*args);
	}//if
	
	gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
	
	g_string_free(rulename,TRUE);
	
	return GAIM_CMD_RET_OK;
}//command_ri


// Regex filter - adds a filter regex 
// for IM and chat messages to the Default level
// /rf expression
static GaimCmdRet	command_rf(GaimConversation *gc, const gchar **args){
	gchar	msgbuf[BUFSIZ];
	ignorance_level	*level;
	ignorance_rule	*rule;
	GString	*rulename;
	
	if(!(*args)){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	level=ignorance_get_default_level();
	
	if(!level){
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to get default level\n");
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	rulename=g_string_new(*args);
	
	rule=ignorance_rule_newp(rulename,IGNORANCE_RULE_REGEX,*args,IGNORANCE_FLAG_FILTER,IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_IM,TRUE,NULL,NULL,NULL);
	
	if(ignorance_level_add_rule(level,rule)){
		snprintf(msgbuf,BUFSIZ,"Successfully added filter regex %s",*args);
	}else{
		snprintf(msgbuf,BUFSIZ,"Unable to add filter regex %s",*args);
	}//if

	gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
	
	g_string_free(rulename,TRUE);
	
	return GAIM_CMD_RET_OK;
}//command_rf


// Regex ignore name - adds an ignore regex 
// for usernames to the Default level
// /rin expression
static GaimCmdRet	command_rin(GaimConversation *gc, const gchar **args){
	gchar	msgbuf[BUFSIZ];
	ignorance_level	*level;
	ignorance_rule	*rule;
	GString	*rulename;
	
	if(!(*args)){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	level=ignorance_get_default_level();
	
	if(!level){
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to get default level\n");
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	rulename=g_string_new(*args);
	
	rule=ignorance_rule_newp(rulename,IGNORANCE_RULE_REGEX,*args,IGNORANCE_FLAG_IGNORE,IGNORANCE_APPLY_USER,TRUE,NULL,NULL,NULL);
	if(ignorance_level_add_rule(level,rule)){
		snprintf(msgbuf,BUFSIZ,"Successfully added ignore regex %s",*args);
	}else{
		snprintf(msgbuf,BUFSIZ,"Unable to add ignore regex %s",*args);
	}//if
	
	gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
	
	g_string_free(rulename,TRUE);
	
	return GAIM_CMD_RET_OK;
}//command_rin


// Regex filter name - adds a filter regex 
// for usernames to the Default level
// /rf expression
static GaimCmdRet	command_rfn(GaimConversation *gc, const gchar **args){
	gchar	msgbuf[BUFSIZ];
	ignorance_level	*level;
	ignorance_rule	*rule;
	GString	*rulename;
	
	if(!(*args)){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	level=ignorance_get_default_level();
	
	if(!level){
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to get default level\n");
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	rulename=g_string_new(*args);
	
	rule=ignorance_rule_newp(rulename,IGNORANCE_RULE_REGEX,*args,IGNORANCE_FLAG_FILTER,IGNORANCE_APPLY_USER,TRUE,NULL,NULL,NULL);
	
	if(ignorance_level_add_rule(level,rule)){
		snprintf(msgbuf,BUFSIZ,"Successfully added filter regex %s",*args);
	}else{
		snprintf(msgbuf,BUFSIZ,"Unable to add filter regex %s",*args);
	}//if

	gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
	
	g_string_free(rulename,TRUE);
	
	return GAIM_CMD_RET_OK;
}//command_rfn


// Searches for a user in specified groups
// /grep user group1 group2 ...
static GaimCmdRet command_grep(GaimConversation *gc, const gchar **args){
	gchar	msgbuf[BUFSIZ],
			*regex=(gchar*)(*args);
	ignorance_level	*level;
	GString	*username,
			*levelname;
	gint	i=1,
			j=0,
			found=0;	
	GList	*users=NULL,
			*cursor;
	gboolean count_only=FALSE;
	
	if(!(args[0])){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	if(!(args[1])){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	levelname=g_string_new("");

	if(!strcmp(args[0],"-c")){
		count_only=TRUE;
		regex=(gchar*)(args[1]);
		++i;
	}//if
	
	for(;args[i];++i){
		found=0;
		g_string_assign(levelname,args[i]);
		gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Searching level %s\n",levelname->str);
		level=ignorance_get_level_name(levelname);
		if(level){
			if(ignorance_level_has_denizen_regex(level,regex,&users)){
				for(cursor=g_list_first(users);cursor;cursor=g_list_next(cursor)){
					username=(GString*)(cursor->data);
					if(!count_only){
						ignorance_print_user_found(gc,username->str,levelname->str);
						//snprintf(msgbuf,BUFSIZ,"Found user %s in level %s",username->str,level->name->str);
						//gaim_conversation_write(gc,NULL,msgbuf,GAIM_MESSAGE_NO_LOG,time(NULL));
					}//if
					++found;
					//(*msgbuf)='\0';
					g_string_free(username,TRUE);
				}//for
				g_list_free(users);

				if(count_only){
					ignorance_print_user_count(gc,regex,levelname->str,found);
					//snprintf(msgbuf,BUFSIZ,"Found user %s in level %s",username->str,level->name->str);
					//gaim_conversation_write(gc,NULL,msgbuf,GAIM_MESSAGE_NO_LOG,time(NULL));
				}//if
			}//if
		}//if level exists
		else if(!strcmp(levelname->str,"*")){
			for(j=0;j<levels->len;++j){
				found=0;
				level=(ignorance_level*)g_ptr_array_index(levels,j);
				users=NULL;
				if(ignorance_level_has_denizen_regex(level,regex,&users)){
					for(cursor=g_list_first(users);cursor;cursor=g_list_next(cursor)){
						username=(GString*)(cursor->data);
						if(!count_only){
							ignorance_print_user_found(gc,username->str,level->name->str);
						}//if
						++found;
						//(*msgbuf)='\0';
						g_string_free(username,TRUE);
					}//for
					g_list_free(users);
					if(count_only){
						ignorance_print_user_count(gc,regex,level->name->str,found);
					}//if
				}//if
				else{
				}//else
			}//for
		}// /grep blah *
	}//for
	
	if(!found){
		snprintf(msgbuf,BUFSIZ,"No match for %s",regex);
		gaim_conversation_write(gc,NULL,msgbuf,GAIM_MESSAGE_NO_LOG,time(NULL));
	}//if
	
	g_string_free(levelname,TRUE);
	
	return GAIM_CMD_RET_OK;
}//command_grep
#endif


// Searches for currently present users in the given groups
// /igwho group1 group2 ...
static GaimCmdRet command_igwho(GaimConversation *gc, const gchar **args){
	GaimConvChat *chat;
	GString	*levelname,
		*username,
		*chatter_name;
	ignorance_level	*level;
	GList	*cursor,
		*chatters,
		*chatter_cursor;
	gchar	*msgbuf;
	gint	i;
	
	if(!args[0] || !(GAIM_CONV_TYPE_CHAT == gaim_conversation_get_type(gc))){
		return GAIM_CMD_RET_CONTINUE;
	}//if

				
	levelname=g_string_new("");
				
	for(i=0;args[i];++i){
		g_string_assign(levelname,args[i]);
		level=ignorance_get_level_name(levelname);
		if(level){
			g_string_append(levelname,":");
			gaim_conversation_write(gc,NULL,levelname->str,GAIM_MESSAGE_NO_LOG,time(NULL));
			chat=gaim_conversation_get_chat_data(gc);
			chatters=gaim_conv_chat_get_users(chat);
			for(cursor=g_list_first(level->denizens);cursor;cursor=g_list_next(cursor)){
				username=(GString*)(cursor->data);
				for(chatter_cursor=chatters;chatter_cursor;chatter_cursor=g_list_next(chatter_cursor)){
					chatter_name=g_string_new(gc->account->protocol_id);
					g_string_append(chatter_name,((GaimConvChatBuddy*)chatter_cursor->data)->name);
					if(g_string_equal(chatter_name,username)){
						gaim_conversation_write(gc,NULL,((GaimConvChatBuddy*)chatter_cursor->data)->name,GAIM_MESSAGE_NO_LOG,time(NULL));
						g_string_free(chatter_name,TRUE);
						break;
					}else{
						g_string_free(chatter_name,TRUE);
					}//if
				}//for each chatter in room
			}//for each chatter in level
		}else{
			msgbuf=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));
			snprintf(msgbuf,BUFSIZ,"No such group: %s",levelname->str);
			gaim_conversation_write(gc,NULL,msgbuf,GAIM_MESSAGE_NO_LOG,time(NULL));
			g_free(msgbuf);
		}//if
	}//for
				
	g_string_free(levelname,TRUE);
	return GAIM_CMD_RET_OK;
}//command_igwho


//
// substitute
// checks an incoming message for rule violation
// account: Account from which the message is being sent
// sender: the name of the sender
// message: Pointer to the message string
// msgflags: message type flags, from ignorance.h
//
// Returns: FALSE if further handling required,
//			TRUE otherwise
static gboolean substitute (GaimConversation *conv,GaimAccount *account, const gchar *sender, gchar **message, gint msgflags){
	GString	*username=NULL,
			*text=NULL;
	gboolean	rv=FALSE;
	gchar	*newmsg,
			*cursor;

	if (NULL == message){
		return FALSE;
	}else if (NULL == (*message)){
		return FALSE;
	}//if

	gaim_debug(GAIM_DEBUG_INFO,"ignorance","Got message \"%s\" from user \"%s\"\n",*message,sender);
	
	username=g_string_new(gaim_normalize(account,sender));
	
	//strip yahoo "tattoo"
	cursor=yahoo_strip_tattoo(*message);
	if(cursor!=*message){
		newmsg=g_strdup(cursor);
		g_free(*message);
		(*message)=newmsg;
	}

	text=g_string_new(*message);
	
	rv=apply_rule(conv,account,username,text,msgflags);
	
	if(rv){
		gaim_debug(GAIM_DEBUG_INFO, "ignorance", "%s: %s violated!\n",username->str,text->str);
	}//if

	g_string_free(username,TRUE);
	g_string_free(text,TRUE);

	gaim_debug(GAIM_DEBUG_INFO,"ignorance","Returning from substitution\n");

	return rv;
}//substitute


// loads the list of levels, rules, and denizens
// from ~/.gaim/ignorance
//
// also tries to load ~/.zinc/ignore and ~/.zinc/whitelist
//
// also imports gaim buddy list into WL
//
// returns success/failure
 static gboolean load_conf(){
	 gchar *buf, 
	       *ibuf;
	 gint pnt = 0;
	 gsize size;
	 FILE *conffile=NULL;
	 static ignorance_level *tmplvl=NULL;
	 static ignorance_rule *tmprule=NULL;
	 GString *tmpgs=NULL;
	
	 buf = g_build_filename(gaim_user_dir(), "ignorance", NULL);
	
	 gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Attempting to load conf file %s\n",buf);
	 
  	levels=g_ptr_array_new();

	 if(!(conffile=fopen(buf,"r"))){
		g_free(buf);
		buf=g_build_filename(IGNORANCE_CONFDIR,"ignorance.conf",NULL);

		if(!(conffile=fopen(buf,"r"))){
			gaim_debug(GAIM_DEBUG_INFO,"ignorance","Unable to open local or global conf files; falling back to defaults\n");
			generate_default_levels();
			import_gaim_list();
			import_zinc_list();

			g_free(buf);
			return FALSE;
		}//default conf
	 }//if
 	 
	 g_file_get_contents(buf, &ibuf, &size, NULL);
	 fclose(conffile);
	 g_free(buf);

	 if(!ibuf) {
		generate_default_levels();
		import_gaim_list();
		import_zinc_list();
		return FALSE;
	 }//if

	while(buf_get_line(ibuf, &buf, &pnt, size)) {
		if((*buf)=='#'){
			 //comment; ignore
		}else if(strstr(buf,"level")){
			if('\0' == buf[5]){	//new style
				for(tmpgs=g_string_new("");buf!=strstr(buf,"/level");buf_get_line(ibuf,&buf,&pnt,size)){
					g_string_append(tmpgs,buf);
					g_string_append(tmpgs,"\n");
				}//read entire level into buf
				tmplvl=ignorance_level_read(tmpgs->str);
				g_string_free(tmpgs,TRUE);
			}else{	//old style
				tmplvl=ignorance_level_read(buf);
			}//if
			if(tmplvl){
	 			gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Adding level %s\n",tmplvl->name->str);
				//g_ptr_array_add(levels,tmplvl);
				ignorance_add_level(tmplvl);
			}//valid level
		 }else if(strstr(buf,"rule") && tmplvl){
			 if('\0' == buf[4]){	//new style
				 for(tmpgs=g_string_new("");buf!=strstr(buf,"/rule");buf_get_line(ibuf,&buf,&pnt,size)){
					g_string_append(tmpgs,buf);
					g_string_append(tmpgs,"\n");
				 }//read entire rule into buf
		 		gaim_debug(GAIM_DEBUG_INFO,"ignorance","Attempting to read rule %s\n",tmpgs->str);
				tmprule=ignorance_rule_read(tmpgs->str);
				g_string_free(tmpgs,TRUE);
			 }else{	//old style
				 tmprule=ignorance_rule_read(buf);
			 }//if
			if(tmprule){
	 			gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Adding rule %s: %s\n",tmprule->name->str,(gchar*)(tmprule->value));
		 		ignorance_level_add_rule(tmplvl,tmprule);
			}//valid rule
		 }else if(tmplvl){//users
			tmpgs=g_string_new(buf);
	 		gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Adding denizen %s\n",buf);
			if(ignorance_get_user_level(tmpgs) == ignorance_get_default_level()){
				ignorance_level_add_denizen(tmplvl,tmpgs);
			}//if
			g_string_free(tmpgs,TRUE);
		}//ugly switch
	}//while

	g_free(ibuf);
	 
	import_gaim_list();
	import_zinc_list();
	
	return TRUE;
 }//load_conf
 

// Generates default groups and rules
static gboolean generate_default_levels(){
	ignorance_level *tmplvl;
	ignorance_rule *tmprule;
	GString	*tmpgs;
	 
 	tmplvl=ignorance_level_new();
 	tmplvl->name=g_string_new("Default");
	ignorance_add_level(tmplvl);
	 
	tmplvl=ignorance_level_new();
	tmplvl->name=g_string_new("WL");
	ignorance_add_level(tmplvl);
	 
 	tmplvl=ignorance_level_new();
	tmplvl->name=g_string_new("BL");
	ignorance_add_level(tmplvl);
	 
	tmpgs=g_string_new("Everything");
#ifdef HAVE_REGEX_H
	tmprule=ignorance_rule_newp(tmpgs,IGNORANCE_RULE_REGEX,".*",IGNORANCE_FLAG_FILTER,IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_IM,TRUE,NULL,NULL,NULL);
#else
	tmprule=ignorance_rule_newp(tmpgs,IGNORANCE_RULE_SIMPLETEXT,"",IGNORANCE_FLAG_FILTER,IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_IM,TRUE,NULL,NULL,NULL);
#endif
	ignorance_level_add_rule(tmplvl,tmprule);
	g_string_free(tmpgs,TRUE);

	return TRUE;
 }//generate_default_levels


//writes the current configuration to ~/.gaim/ignorance
gboolean save_conf()
{
	FILE *f;
	gchar	*name,
			tempfilename[BUF_LONG];
	gint fd,
		i;

	name = g_build_filename(gaim_user_dir(), "ignorance", NULL);
	strcpy(tempfilename, name);
	strcat(tempfilename,".XXXXXX");
	fd = g_mkstemp(tempfilename);
	if(fd<0) {
		perror(tempfilename);
		g_free(name);
		return FALSE;
	}//if
	if (!(f = fdopen(fd, "w"))) {
		perror("fdopen");
		close(fd);
		g_free(name);
		return FALSE;
	}//if

	fchmod(fd, S_IRUSR | S_IWUSR);

	for(i=0;i<levels->len;++i){
		ignorance_level_write(g_ptr_array_index(levels,i),f);
	}//for
	
	if(fclose(f)) {
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance",
				   "Error writing to %s: %m\n", tempfilename);
		unlink(tempfilename);
		g_free(name);
		return FALSE;
	}//if
	rename(tempfilename, name);
	g_free(name);
	return TRUE;
}//save_conf


// gets a line from a buffer
//
// ibuf is the buffer to be searched
// buf will be pointed to the beginning of the next line
// position marks the current position in ibuf
// len is the length of ibuf
//
// returns true if there is more to read, false otherwise
static gint buf_get_line(gchar *ibuf, gchar **buf, gint *position, gint len) {
	gint pos = *position, spos = pos;

	if (pos == len)
		return 0;

	while (ibuf[pos++] != '\n') {
		if (pos == len)
			return 0;
	}
	pos--;
	ibuf[pos] = 0;
	*buf = &ibuf[spos];
	pos++;
	*position = pos;
	return 1;
}//buf_get_line


//checks a given level for a given user
static gboolean	ignorance_user_match(ignorance_level *il, const GString *username){
	return (ignorance_level_has_denizen(il,username));
}//ignorance_user_match


//returns the index of the level to which a user belongs, or IGNORANCE_LEVEL_DEFAULT
static ignorance_level* ignorance_get_user_level(const GString *username){
	gint i=0;
	ignorance_level	*il=NULL;

	for(i=0;i<levels->len;++i){
		il=(ignorance_level*)g_ptr_array_index(levels,i);
		if(ignorance_user_match(il,username)){
			return il;
			//break;
		}//if
	}//for

	//wtf?
	//if(i<levels->len){
	//	return il;
	//}//if

	return ignorance_get_default_level();
}//ignorance_get_user_level


// returns a pointer to the default group
static ignorance_level* ignorance_get_default_level(){
	gint i=0;
	ignorance_level *il;

	for(i=0;i<levels->len;++i){
		il=(ignorance_level*)g_ptr_array_index(levels,i);
		if(!strcmp(il->name->str,"Default")){
			return il;
		}//if
	}//for

	return NULL;
}//ignorance_get_default_level


//returns a level specified by name, or NULL
ignorance_level*	ignorance_get_level_name(const GString *levelname){
	gint i=0;
	ignorance_level *il=NULL;

	for(i=0;i<levels->len;++i){
		il=(ignorance_level*)g_ptr_array_index(levels,i);
		if(g_string_equal(levelname,il->name)){
			return il;
		}//if
	}//for

	return NULL;
}//ignorance_get_level_name


//places a given user at a given level, removing him from his current level and returning success/failure
static gboolean	ignorance_place_user_name(const GString *level_name, const GString *username){
	ignorance_level	*current_level=ignorance_get_user_level(username),
					*newlevel=ignorance_get_level_name(level_name);

	if(newlevel){
		if(newlevel!=current_level){
			ignorance_level_remove_denizen(current_level,username);
			ignorance_level_add_denizen(newlevel,username);	
			return TRUE;
		}//if
	}//if
	else{
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Invalid level %s\n",level_name->str);
	}//else
	
	return FALSE;
}//ignorance_place_user_name


// applies a user's message to the rules for his current level
//
// username is the user
// text is the message
// flags are the message type flags
//
// returns true if ok, false on violation
static gboolean	apply_rule(GaimConversation *conv,GaimAccount *account,const GString *username, const GString *text, gint flags){
	ignorance_level	*user_level;
	gint	text_score=0;
	gboolean	rv=TRUE,
				newconv=FALSE;
	GList	*violations=NULL,
		 	*cursor=NULL;
	ignorance_violation *viol;
	GString *tmp;
	GString *prpluser=g_string_new(gaim_account_get_protocol_id(account));

	if(strstr(prpluser->str,"prpl-jabber")==prpluser->str){
	}//if jabber

	g_string_append(prpluser,gaim_normalize(account,username->str));
	user_level=ignorance_get_user_level(prpluser);

	gaim_debug(GAIM_DEBUG_INFO,"ignorance","Preparing to check %s\n",text->str);
	
	text_score=ignorance_level_rulecheck(user_level,prpluser,text,flags,&violations);

	gaim_debug(GAIM_DEBUG_INFO,"ignorance","Got score %d\n",text_score);

	if(!(text_score & (IGNORANCE_FLAG_FILTER | IGNORANCE_FLAG_IGNORE))){
		rv=FALSE;
		if(text_score){
			for(cursor=violations;cursor;cursor=cursor->next){
				viol=(ignorance_violation*)(cursor->data);
				gaim_debug(GAIM_DEBUG_INFO,"ignorance","Got violation type %d: %s\n",viol->type,viol->value);
				switch(viol->type){
					case IGNORANCE_FLAG_SOUND:
						gaim_debug(GAIM_DEBUG_INFO,"ignorance","Attempting to play sound %s\n",viol->value);
#if ((GAIM_MAJOR_VERSION) == 2)
						gaim_sound_play_file(viol->value,account);
#else
						gaim_sound_play_file(viol->value);
#endif
						break;
#ifndef _WIN32
					case IGNORANCE_FLAG_EXECUTE:
						gaim_debug(GAIM_DEBUG_INFO,"ignorance","Attempting to execute command %s\n",viol->value);
						tmp=g_string_new("");
						handle_exec_command(viol->value,tmp,512);

						if(conv){
								gaim_conversation_write(conv,gaim_account_get_username(account),tmp->str,GAIM_MESSAGE_NO_LOG,time(NULL));
						}//if existing conversation

						g_string_free(tmp,TRUE);
						break;
#endif
					case IGNORANCE_FLAG_MESSAGE:
						if(!conv){
							newconv=TRUE;
							conv=gaim_conversation_new(GAIM_CONV_TYPE_IM,account,username->str);
						}//no existing conversation with sender

						switch(gaim_conversation_get_type(conv)){
							case GAIM_CONV_TYPE_IM:
								gaim_conv_im_send(GAIM_CONV_IM(conv),viol->value);
								break;
							case GAIM_CONV_TYPE_CHAT:
								gaim_conv_chat_send(GAIM_CONV_CHAT(conv),viol->value);
								break;
							default:
								break;	//do nothing
						}//check type

						if(newconv){
							gaim_conversation_destroy(conv);
						}//destroy newly created conv
						break;
					default:
						break;
				}//select
			}//process violations
		}//message/execute/sound hits
	}//no filter/ignore hits
	else if(text_score & IGNORANCE_FLAG_IGNORE){
		//tmp=g_string_new("BL");
		//ignorance_place_user_name(tmp,username);
		//g_string_free(tmp,TRUE);
		ignorance_bl_user(conv,username->str,"BL");
	}//ignore user

	g_string_free(prpluser,TRUE);

	gaim_debug(GAIM_DEBUG_INFO,"ignorance","Preparing to free violation items\n");
	g_list_foreach(violations,ignorance_violation_free_g,NULL);
	gaim_debug(GAIM_DEBUG_INFO,"ignorance","Done freeing violation items, now freeing list itself\n");
	g_list_free(violations);

	gaim_debug(GAIM_DEBUG_INFO,"ignorance","Done checking, returning from applying rules\n");

	return rv;
}//apply_rule


//attemps to import ignore list and whitelist from ~/.zinc
static gboolean import_zinc_list(){
	gchar	*buf,
			*ibuf;
	gint	pnt=0;
	gsize	size=0;
	FILE	*zincfile;
	ignorance_level	*zinclevel=NULL;
	ignorance_rule	*zincrule=NULL;
	GString	*tmp=NULL;
	
	buf=g_build_filename(g_get_home_dir(),".zinc","ignore", NULL);
	
	if(!(zincfile=fopen(buf,"r"))){
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to open %s\n",buf);
		g_free(buf);
		return FALSE;
	}else{
		fclose(zincfile);
	}//if

	
	g_file_get_contents(buf, &ibuf, &size, NULL);
	
	tmp=g_string_new("ZincBL");
	zinclevel=ignorance_get_level_name(tmp);
	if(!zinclevel){
		gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Creating new Zinc blacklist\n");
#ifdef HAVE_REGEX_H
		zincrule=ignorance_rule_newp(g_string_new("Everything"),IGNORANCE_RULE_REGEX,(gchar*)".*",IGNORANCE_FLAG_FILTER,(IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_IM | IGNORANCE_APPLY_ENTERLEAVE),TRUE,NULL,NULL,NULL);
#else
		zincrule=ignorance_rule_newp(g_string_new("Everything"),IGNORANCE_RULE_SIMPLETEXT,(gchar*)"",IGNORANCE_FLAG_FILTER,(IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_IM | IGNORANCE_APPLY_ENTERLEAVE),TRUE,NULL,NULL,NULL);
#endif
				
		zinclevel=ignorance_level_new();
		zinclevel->name=g_string_new(tmp->str);
		ignorance_level_add_rule(zinclevel,zincrule);
		ignorance_add_level(zinclevel);
	}//if no ZincBL already
	
	if(!tmp){
		tmp=g_string_new("");
	}//if

	g_free(buf);
	
	gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Preparing to read in zinc blacklist users\n");
	for(;buf_get_line(ibuf, &buf, &pnt, size);){
		g_string_assign(tmp,"prpl-yahoo");
		g_string_append(tmp,buf);
		if(ignorance_get_user_level(tmp) == ignorance_get_default_level()){
			ignorance_level_add_denizen(zinclevel,tmp);
		}//if
	}//for

	g_free(ibuf);
	
	pnt=0;
	buf=g_build_filename(g_get_home_dir(),".zinc","whitelist", NULL);
	
	g_file_get_contents(buf, &ibuf, &size, NULL);
	
	g_string_assign(tmp,"ZincWL");
	zinclevel=ignorance_get_level_name(tmp);
	if(!zinclevel){
		gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Creating new Zinc whitelist\n");
		zinclevel=ignorance_level_new();
		zinclevel->name=g_string_new(tmp->str);
		ignorance_add_level(zinclevel);
	}//if no ZincWL already

	g_free(buf);
	
	gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Preparing to read in zinc whitelist users\n");
	for(;buf_get_line(ibuf, &buf, &pnt, size);){
		g_string_assign(tmp,"prpl-yahoo");
		g_string_append(tmp,buf);
		if(ignorance_get_user_level(tmp) == ignorance_get_default_level()){
			ignorance_level_add_denizen(zinclevel,tmp);
		}//if
	}//for

	g_free(ibuf);
	
	gaim_debug(GAIM_DEBUG_INFO, "ignorance", "Done importing Zinc users\n");
	
	return TRUE;
}//import_zinc_list


static gboolean import_gaim_list(){
	GaimBuddyList *bl=gaim_get_blist();	
	GString	*wlname=g_string_new("WL");
	ignorance_level	*wl=ignorance_get_level_name(wlname);
	gboolean	rv=FALSE;
	
	if(bl && wl){
		g_hash_table_foreach(bl->buddies,gaim_buddy_add,wl);
		rv=TRUE;
	}else{
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to get Gaim buddy list!\n");
	}//if
	
	g_string_free(wlname,TRUE);
	
	return rv;
}//import_gaim_list


// key is unused
// value is the GaimBuddy* to be added
// user_data is the ignorance_level* to add to
static void gaim_buddy_add(gpointer key, gpointer value, gpointer user_data){
	GaimBuddy	*buddy=(GaimBuddy*)value;
	ignorance_level	*level=(ignorance_level*)user_data;
	GString	*tmp;
	
	if(buddy && level){
		tmp=g_string_new(buddy->account->protocol_id);
		g_string_append(tmp,gaim_normalize(buddy->account,buddy->name));
		
		if(ignorance_get_user_level(tmp) == ignorance_get_default_level()){
			ignorance_level_add_denizen(level,tmp);

			//HACK!
			if(strstr(level->name->str,"WL")){
				//work with internal gaim privacy, such as it is
				gaim_privacy_deny_remove(buddy->account,
											buddy->name,FALSE);
				gaim_privacy_permit_add(buddy->account,
										buddy->name, FALSE);
			}else if(strstr(level->name->str,"BL")){
				//work with internal gaim privacy, such as it is
				gaim_privacy_permit_remove(buddy->account,
											buddy->name,FALSE);
				gaim_privacy_deny_add(buddy->account,
										buddy->name, FALSE);
			}//if WL or BL
		}//if
		
		g_string_free(tmp,TRUE);
	}else{
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Bad arguments to gaim_buddy_add\n");
	}//if
}//gaim_buddy_add


static gboolean ignorance_wl_user(GaimConversation *gc, const gchar *username, const gchar *actual_levelname){
	gboolean rv=FALSE;
	GString *wlname=g_string_new(actual_levelname),
			*usergs=g_string_new(gc->account->protocol_id);
	gchar *msgbuf=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));

	g_string_append(usergs,username);

	if(ignorance_place_user_name(wlname,usergs)){
		rv=TRUE;

		//work with internal gaim privacy, such as it is
		gaim_privacy_deny_remove(gaim_conversation_get_account(gc),
									username,FALSE);
		gaim_privacy_permit_add(gaim_conversation_get_account(gc),
								username, FALSE);
		if(gaim_conversation_get_type(gc) == GAIM_CONV_TYPE_CHAT){
			gaim_conv_chat_unignore(GAIM_CONV_CHAT(gc),username);
		}//unignore in chat too - I don't think this is redundant with gaim privacy

		snprintf(msgbuf,BUFSIZ,"Assigned user %s to %s",username,actual_levelname);
	}else{
		snprintf(msgbuf,BUFSIZ,"Unable to assign user %s to %s - may already be there",username,actual_levelname);
	}//if

	gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));

	g_free(msgbuf);
	g_string_free(usergs,TRUE);
	g_string_free(wlname,TRUE);

	return rv;
}//ignorance_wl_user


static gboolean ignorance_rm_user(GaimConversation *gc, const gchar *username){
	GString	*usergs=g_string_new(username);	//should already have protocol prepended
	ignorance_level	*level;
	gchar	*msgbuf;
	gboolean rv=FALSE;
	
	level=ignorance_get_user_level(usergs);
	if(level){
		rv=ignorance_level_remove_denizen(level,usergs);

		if(gc){
			//work with internal gaim privacy, such as it is
			gaim_privacy_permit_remove(gaim_conversation_get_account(gc),
										username,FALSE);
			gaim_privacy_deny_remove(gaim_conversation_get_account(gc),
										username,FALSE);
			if(gaim_conversation_get_type(gc) == GAIM_CONV_TYPE_CHAT){
				gaim_conv_chat_unignore(GAIM_CONV_CHAT(gc),username);
			}//unignore in chat too - I don't think this is redundant with gaim privacy
				
			msgbuf=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));

			if(rv){
				snprintf(msgbuf,BUFSIZ,"Successfully removed %s from %s",username,level->name->str);
				rv=TRUE;
			}else{
				snprintf(msgbuf,BUFSIZ,"Unable to remove %s from %s\n",username,level->name->str);
			}//if

			gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
			g_free(msgbuf);
		}//if
	}//if
				
	g_string_free(usergs,TRUE);

	return rv;
}//ignorance_rm_user


static gboolean ignorance_bl_user(GaimConversation *gc, const gchar *username, const gchar *actual_levelname){
	gboolean rv=FALSE;
	GString *wlname=g_string_new(actual_levelname),
			*usergs=g_string_new(gc->account->protocol_id);
	gchar *msgbuf=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));

	g_string_append(usergs,username);

	if(ignorance_place_user_name(wlname,usergs)){
		rv=TRUE;
		
		//work with internal gaim privacy, such as it is
		gaim_privacy_permit_remove(gaim_conversation_get_account(gc),
									username,FALSE);
		gaim_privacy_deny_add(gaim_conversation_get_account(gc),
								username, FALSE);
		if(gaim_conversation_get_type(gc) == GAIM_CONV_TYPE_CHAT){
			gaim_conv_chat_ignore(GAIM_CONV_CHAT(gc),username);
		}//unignore in chat too - I don't think this is redundant with gaim privacy

		snprintf(msgbuf,BUFSIZ,"Assigned user %s to %s",username,actual_levelname);
	}else{
		snprintf(msgbuf,BUFSIZ,"Unable to assign user %s to %s - may already be there",username,actual_levelname);
	}//if

	gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));

	g_free(msgbuf);
	g_string_free(usergs,TRUE);
	g_string_free(wlname,TRUE);

	return rv;
}//ignorance_bl_user


/*
 * strips "tattoo" added by some yahoo clients
 * "tattoo" generally defined as <font name="some wingding font" tattoo>stuff</font> at beginning of message
 * returns a pointer to the first non-tattoo char in message,
 * or message itself if no tattoo detected
 */
static gchar* yahoo_strip_tattoo(gchar *origmessage){
	gchar	*cursor,
			*cursor2,
			*beginstack,
			*endstack,
			*message = g_ascii_strdown(origmessage,-1);
	gint    rvindex;

	cursor = strstr(message,"<font");

	if(message == cursor){
		cursor2 = strstr(message,"tattoo");
		if(cursor2){
			cursor = strstr(cursor2,">"); // end of first font tag

			for(;cursor;){
				endstack = strstr(cursor,"</");
				beginstack = strstr(cursor,"<");

				if(NULL == endstack || NULL == beginstack){
					cursor = NULL;
					break;
				}//no more tags

				cursor = strstr(endstack,">"); // (hopefully) set cursor at end of inner tag 
				if(beginstack == endstack){
					break;
				}// end of outer tag
			}//for(;cursor;)

			if(cursor){
				rvindex = cursor+1-message;
				gaim_debug(GAIM_DEBUG_INFO, "yahoo",
					"%s\nconverted to \n%s\n%s\n\n",
					origmessage,cursor+1,origmessage+rvindex);
				g_free(message);
				return origmessage+rvindex;
			}//if(cursor)
		}//if(cursor2)
	}//if(message == cursor)

	g_free(message);
	return origmessage; // no tattoo found
}//yahoo_strip_tattoo


#ifndef _WIN32
/*
 * handle_exec_command
 * Runs a sh command and stores the output in a GString
 * Param command: The command to be run
 * Param result: A string where the command output will be stored
 * Param maxlen: The maximum acceptable output length
 * Returns: Currently always 0
 */
static int
handle_exec_command (const gchar *command, GString *result,unsigned long maxlen){
	GString *inp=g_string_new("");
			
	int p[2], pid, chrs;

	pipe(p);

	chrs = inp->len;
	if ((pid = fork ()) == -1){	//Error forking
		g_string_assign (result, command);
		g_string_append (result, ": couldn't fork");
		return -1;
    }else if (pid){		//Parent
		int rv;

		close (p[1]);
		int flags = fcntl (p[0], F_GETFL, 0);
		fcntl (p[0], F_SETFL, flags | O_NONBLOCK);
		
		rv=read_nonblock(p[0],maxlen-chrs,EXEC_TIMEOUT,inp);	

	    if (kill (pid, 0) == 0){	//kill child after reading max allowed output
				 kill (pid, SIGKILL);
		}//if
		if (rv<0){
			g_string_append (inp, "[process timed out]");
	    }else if(maxlen==(rv+chrs)){	//Output was truncated
			g_string_append(inp,"...");
	    }else if(('\n'==(inp->str)[inp->len-1])){
		    g_string_truncate(inp,inp->len-1);
		}//if

		g_string_assign (result, inp->str);
		
		g_string_free(inp,TRUE);
		waitpid (pid, NULL, 0);
	}else{	//Child
	    close (p[0]);
		if (p[1] != STDOUT_FILENO){
			dup2 (p[1], STDOUT_FILENO);
			close (p[1]);
		}//if
		if (STDERR_FILENO != STDOUT_FILENO){
			dup2 (STDOUT_FILENO, STDERR_FILENO);
		}//if
		execlp ("sh", "sh", "-c", command, NULL);
	}//if parent/child
	return 0;
}//handle_exec_command


/*
 * read_nonblock
 * Reads up to a specified number of characters from a nonblocking fd,
 * or up to a specified timeout
 * Param fd: The file descriptor for reading
 * Param len: The maximum number of characters to be read
 * Param timeout: The maximum time to wait, in seconds
 * Param inp: The GString* that will receive the input from the fd
 * Returns: The number of characters read, or -1 on timeout
 */
static int read_nonblock(int fd,unsigned long len,unsigned long timeout,GString *inp){
	int chrs=0,
		timedout=FALSE,
		rv=0;
	unsigned long 	timeout_usec=timeout*1000000,
					tlapsed=0;
	static const int SLEEP_INTERVAL=50000;

	gchar	*ptr=g_malloc((len+1)*sizeof(gchar));

	while (chrs < len){
		if (tlapsed > timeout_usec){
			timedout = TRUE;
			break;
		}//if
		if ((rv = read (fd, ptr, (len - chrs))) < 0){	//Error reading - wait and try again
			rv = errno;
			usleep (SLEEP_INTERVAL);
			tlapsed += SLEEP_INTERVAL;
			continue;
		}else if (rv == 0){	//Done reading
			break;
		}//if
		chrs += rv;
		*(ptr + rv)= 0;
		g_string_append(inp,ptr);
		*ptr=0;
	}//while
	g_free(ptr);
	if(timedout){
		return -1;
	}else{
		return chrs;
	}//if
}//read_nonblock
#endif


gboolean ignorance_add_level(ignorance_level *level){
	gboolean rv=FALSE;

	if(level){
		g_ptr_array_add(levels,level);
		rv=TRUE;
	}//if

	return rv;
}//ignorance_add_level


gboolean ignorance_remove_level(const GString *levelname){
	ignorance_level *level=ignorance_get_level_name(levelname);
	gboolean rv=FALSE;

	if(level){
		rv=g_ptr_array_remove(levels,level);
		ignorance_level_free(level);
	}//if

	return rv;
}//ignorance_remove_level


static void ignorance_print_user_found(GaimConversation *conv, const gchar *username, const gchar *levelname){
	gchar *msgbuf=g_strdup_printf("Found user %s in level %s",username,levelname);
	gaim_conversation_write(conv,NULL,msgbuf,GAIM_MESSAGE_NO_LOG,time(NULL));
	g_free(msgbuf);
}//ignorance_print_user_found


static void ignorance_print_user_count(GaimConversation *conv, const gchar *regex, const gchar *levelname, gint count){
	gchar *msgbuf=g_strdup_printf("Found %d users matching %s in level %s",count,regex,levelname);
	gaim_conversation_write(conv,NULL,msgbuf,GAIM_MESSAGE_NO_LOG,time(NULL));
	g_free(msgbuf);
}//ignorance_print_user_count


static GaimCmdRet command_msg(GaimConversation *c, const gchar *cmd, gchar **args, gchar **error, void *data){

	GaimAccount *account;
	GaimConversation *imc;

	if(!args || !args[0])
		return GAIM_CMD_RET_FAILED;

	account = gaim_conversation_get_account(c);

//Damn you, API changes!
#if ((GAIM_MAJOR_VERSION) == 2)
	imc = gaim_find_conversation_with_account(GAIM_CONV_TYPE_IM,args[0],account);
#else
	imc = gaim_find_conversation(args[0]);
#endif

	if(!imc){	// no existing conv, create new
		imc = gaim_conversation_new(GAIM_CONV_TYPE_IM,account,args[0]);
	}

	if(args[1]){
		gaim_conv_im_send(GAIM_CONV_IM(imc), args[1]);
	}//if

	return GAIM_CMD_RET_OK;
}//ignorance_cmd_msg

